

/*
 *  Licensed Materials - Property of IBM
 *  Restricted Material of IBM
 *  IBM Rational Test RealTime
 *  (c) Copyright IBM Corp. 2003-2011 All Rights Reserved.
 *
 *  US Governement Users Restricted Rights - Use, duplication or
 *  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */


/*******************************************************************************
* $Descr: PurifyLT Header (included across TP.h) $
*******************************************************************************/


/* Multiple inclusions guard: */
#ifndef ATP_H_INCLUDED
#define ATP_H_INCLUDED


/*******************************************************************************
*  C/C++ Instrumentors
*******************************************************************************/

/*******************************************************************************
*  Parsing global data function management
*******************************************************************************/

#if !RTRT_KR
typedef void                    (*_AtpParseGlobalsFuncTP) ( void );
#else /* RTRT_KR */
typedef void                    (*_AtpParseGlobalsFuncTP) ( );
#endif /* RTRT_KR */

#if RTRT_CPLUSPLUS_LANGUAGE
extern "C" {
#endif /* RTRT_CPLUSPLUS_LANGUAGE */

typedef struct _AtpGlobalsFuncType_ {
        struct _AtpGlobalsFuncType_ *Next;
        _AtpParseGlobalsFuncTP       ParseGlobalsFunc;
#ifdef ATL_T_FAMILY
        ATL_T_FAMILY				 Family;
#endif
} _AtpGlobalsFuncType, *_AtpGlobalsFuncTypeP;

#ifndef ATL_T_FAMILY
#define _ATP_FAMILY_INIT
#else
#ifndef ATL_FAMILY_INIT_VALUE
#define ATL_FAMILY_INIT_VALUE (ATL_T_FAMILY)0
#endif /*ATL_FAMILY_INIT_VALUE */
#define _ATP_FAMILY_INIT    _atp_globals_func_data.Family=ATL_FAMILY_INIT_VALUE;
#endif /* _ATL_T_FAMILY */

#define _ATP_PARSE_GLOBALS _atp_parse_globals

#if RTRT_CPLUSPLUS_LANGUAGE
#define _ATP_LINK_PARSE_GLOBALS_ \
  if ( RTRT_IS_NOT_INITIALIZED( _atp_parse_globals_init_done) ) \
  { _atp_parse_globals_init_done = RTRT_VARIABLE_INITIALIZED; \
    _atp_globals_func_data.ParseGlobalsFunc = ::_ATP_PARSE_GLOBALS; \
    _ATP_FAMILY_INIT \
    _atp_link_parse_globals(&_atp_globals_func_data); }
#else
#define _ATP_LINK_PARSE_GLOBALS_ \
  if ( RTRT_IS_NOT_INITIALIZED(_atp_parse_globals_init_done )) \
  { _atp_parse_globals_init_done = RTRT_VARIABLE_INITIALIZED; \
    _atp_globals_func_data.ParseGlobalsFunc = _ATP_PARSE_GLOBALS; \
    _ATP_FAMILY_INIT \
    _atp_link_parse_globals(&_atp_globals_func_data); }
#endif

#define _ATP_DEFINE_PARSE_GLOBALS \
static int _atp_parse_globals_init_done ATL_SET_SECTION ATL_OBSTOOLS_INITIALIZE_TO_ZERO; \
static _AtpGlobalsFuncType _atp_globals_func_data ATL_SET_SECTION;


/*******************************************************************************
*  Parsing global data
*******************************************************************************/

typedef struct {
        char           *Addr;
        RTRT_U_INT32    Size;
        RTRT_U_INT8     Used;
} _AtpDeepGlobalType, *_AtpDeepGlobalTypeP;

#if ATL_DYN_TABLE
#ifndef _ATP_ALLOC_TABLE
#define _ATP_ALLOC_TABLE(n,s) priv_malloc(n*s)
#define _ATP_FREE_TABLE(p) priv_free(p)
#endif /* _ATC_ALLOC */

#define _ATP_DEFINE_DEEP_GLOBAL(n) \
static _AtpDeepGlobalType *_atp_deep_global_table ATL_SET_SECTION =0; \
static void _atp_deep_global_table_init(void)   { \
  { int i; \
    _atp_deep_global_table=(_AtpDeepGlobalTypeP)_ATP_ALLOC_TABLE(n,sizeof(_AtpDeepGlobalType)); \
    for (i = 0;i<n;i++)_atp_deep_global_table[i].Used = 0; }}

#define _ATP_USE_DEEP_GLOBAL(v,n) \
  if(_atp_deep_global_table==0)_atp_deep_global_table_init(); \
  _atp_deep_global_table[n].Addr = (char *)&(v); \
  _atp_deep_global_table[n].Size = sizeof(v); \
  _atp_deep_global_table[n].Used = 1;

#define _ATP_PARSE_DEEP_GLOBALS(m) \
  { int n; \
    if(_atp_deep_global_table) for ( n = 0; n < m; n++ ) \
    { \
      if ( _atp_deep_global_table[n].Used ) \
      { \
      _atp_check_var ( _atp_deep_global_table[n].Addr, _atp_deep_global_table[n].Size ); \
      } } }

#else /* ATL_DYN_TABLE */

/* define _ATP_PARSE_GLOBAL(v) allows to disable the glob var checking and so reduce the _aug code */
#if RTRT_HEAP == RTRT_NONE || defined(_ATP_PARSE_GLOBAL)
#define _ATP_DEFINE_DEEP_GLOBAL(n)
#define _ATP_USE_DEEP_GLOBAL(v,n)
#define _ATP_PARSE_DEEP_GLOBALS(m)
#else /*RTRT_HEAP == RTRT_NONE*/

#define _ATP_DEFINE_DEEP_GLOBAL(n) \
static _AtpDeepGlobalType _atp_deep_global_table[n] ATL_SET_SECTION;

#define _ATP_USE_DEEP_GLOBAL(v,n) \
  _atp_deep_global_table[n].Addr = (char *)&(v); \
  _atp_deep_global_table[n].Size = sizeof(v); \
  _atp_deep_global_table[n].Used = 1;

#define _ATP_PARSE_DEEP_GLOBALS(m) \
  { int n; \
    for ( n = 0; n < m; n++ ) \
    { \
      if ( _atp_deep_global_table[n].Used ) \
      { \
      _atp_check_var ( _atp_deep_global_table[n].Addr, _atp_deep_global_table[n].Size ); \
      } } }
#endif /*RTRT_HEAP == RTRT_NONE*/

#endif /* ATL_DYN_TABLE */



#if !RTRT_KR
#define _ATP_START_PARSE_GLOBALS \
static void _ATP_PARSE_GLOBALS ( void ) \
{
#else
#define _ATP_START_PARSE_GLOBALS \
static void _ATP_PARSE_GLOBALS ( ) \
{
#endif

#define _ATP_END_PARSE_GLOBALS \
}

#ifndef _ATP_PARSE_GLOBAL
#if RTRT_HEAP == RTRT_NONE
#define _ATP_PARSE_GLOBAL(v)
#else
#define _ATP_PARSE_GLOBAL(v) _atp_check_var ( (char *)&(v), sizeof ( v ));
#endif /*RTRT_HEAP == RTRT_NONE*/
#endif /*_ATP_PARSE_GLOBAL*/

#define _ATP_MASK_FAR(v) (0x80000000LU|(v))

#if RTRT_CPLUSPLUS_LANGUAGE
} /*  end of extern "C" */
#endif /* RTRT_CPLUSPLUS_LANGUAGE */
#ifdef ATL_NO_STACK
#define MEMPRONOLINENUMBER
#endif


/*******************************************************************************
*  C Instrumentor
*******************************************************************************/
#if ! RTRT_CPLUSPLUS_LANGUAGE

#define _ATP_IN(i) \
  _ATP_LINK_PARSE_GLOBALS ;

#define _ATP_OUT

#define _ATP_OUT_EXPR 0

#ifndef MEMPRONOLINENUMBER
#define _ATP_LINE_OFFSET(i,o) \
  _atl_stack_link.Info.SourceLineOffset = o;
#else
#define _ATP_LINE_OFFSET(i,o)
#endif

#ifndef MEMPRONOLINENUMBER
#define _ATP_LINE_OFFSET_EXPR(i,o) \
  (_atl_stack_link.Info.SourceLineOffset = o, 0)
#else
#define _ATP_LINE_OFFSET_EXPR(i,o) 0
#endif

#ifndef MEMPRONOLINENUMBER
#define _ATP_LINE_OFFSET_EXPR_START(i,o) \
  _atl_stack_link.Info.SourceLineOffset = o
#else
#define _ATP_LINE_OFFSET_EXPR_START(i,o) 0
#endif

#ifndef _ATP_LINE_OFFSET_RETURN
#define _ATP_LINE_OFFSET_RETURN(tsf,line)
#endif
#define _ATP_LINE_OFFSET_BLOCK  _ATP_LINE_OFFSET
#define _ATP_LINE_OFFSET_THROW  _ATP_LINE_OFFSET

#define _ATP_START_OF_MAIN \
  { \
    char _atu_atp_addr; _atp_start_test_driver(&_atu_atp_addr); \
    _ATP_LINK_PARSE_GLOBALS \
  }

#define _ATU_ATP_START_OF_MAIN _ATP_START_OF_MAIN

#define _ATL_DUMP_STACK(name) atl_dump_CALLSTACK(name,&_atl_stack_link);

#endif /* C language */


/*******************************************************************************
*  C++ Instrumentor
*******************************************************************************/
#if RTRT_CPLUSPLUS_LANGUAGE

#define _ATP_IN(i) \
  _ATP_LINK_PARSE_GLOBALS ;

#ifndef MEMPRONOLINENUMBER
#define _ATP_LINE_OFFSET(i,o) \
  _atl_stack_object._atl_stack_link.Info.SourceLineOffset = o;
#else
#define _ATP_LINE_OFFSET(i,o)
#endif

#define _ATP_LINE_OFFSET_DECL(i,o) \
  _ATP_LINE_OFFSET(i,o)

#ifndef MEMPRONOLINENUMBER
#define _ATP_LINE_OFFSET_EXPR_START(i,o) \
  _atl_stack_object._atl_stack_link.Info.SourceLineOffset = o
#else
#define _ATP_LINE_OFFSET_EXPR_START(i,o) 0
#endif

#ifndef _ATP_LINE_OFFSET_RETURN
#define _ATP_LINE_OFFSET_RETURN _ATP_LINE_OFFSET
#endif
#define _ATP_LINE_OFFSET_BLOCK  _ATP_LINE_OFFSET
#define _ATP_LINE_OFFSET_THROW  _ATP_LINE_OFFSET

#define _ATP_START_PARSE_CLASS_GLOBAL \
  public: static void _ATP_PARSE_GLOBALS ( ) {

#define _ATP_DECLARE_FAR_PARSE_CLASS_GLOBAL \
  public: inline static void _ATP_PARSE_GLOBALS ( );

#define _ATP_START_FAR_PARSE_CLASS_GLOBAL(t) \
  void t::_ATP_PARSE_GLOBALS ( ) {

#define _ATP_END_PARSE_CLASS_GLOBAL \
  }

#define _ATP_START_OF_MAIN \
  { \
    char _ato_atp_addr; _atp_start_test_driver(&_ato_atp_addr); \
    _ATP_LINK_PARSE_GLOBALS \
  }

#define _ATO_ATP_START_OF_MAIN _ATP_START_OF_MAIN

#define _ATL_DUMP_STACK(name) atl_dump_CALLSTACK(name,&_atl_stack_object._atl_stack_link);

#endif /* C++ language */

/*******************************************************************************
* Library Allocation Functions:
*******************************************************************************/

#define ATP_LIBC_PROTO_PREFIX
#if ( RTRT_HEAP == RTRT_STD ) && ( RTRT_LOWLEVEL_MEMLIB == RTRT_NONE )
/* This is the preprocessing way to catch the calls */
#ifndef RTRT_TDP_COMPILATION
#define malloc      _atp_malloc
#define calloc      _atp_calloc
#define realloc     _atp_realloc
#define free        _atp_free
#if RTRT_ATP_STRDUP
#define strdup      _atp_strdup
#endif
#else /* RTRT_TDP_COMPILATION */
#ifndef RTRT_DO_MALLOC /* To be redef in custom.h */
#define RTRT_DO_MALLOC     malloc
#endif /* RTRT_DO_MALLOC */
#ifndef RTRT_DO_CALLOC /* To be redef in custom.h */
#define RTRT_DO_CALLOC     calloc
#endif /* RTRT_DO_CALLOC */
#ifndef RTRT_DO_REALLOC /* To be redef in custom.h */
#define RTRT_DO_REALLOC    realloc
#endif /* RTRT_DO_REALLOC */
#ifndef RTRT_DO_FREE /* To be redef in custom.h */
#define RTRT_DO_FREE       free
#endif /* RTRT_DO_FREE */
ATL_NAMESPACE_BEGIN
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX void * ATL_CALLCONV _atp_malloc  ( RTRT_SIZE_T ) RTRT_THROWING_EXCEPTION;
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX void * ATL_CALLCONV _atp_calloc  ( RTRT_SIZE_T, RTRT_SIZE_T ) RTRT_THROWING_EXCEPTION;
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX void * ATL_CALLCONV _atp_realloc ( void *, RTRT_SIZE_T ) RTRT_THROWING_EXCEPTION;
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX void   ATL_CALLCONV _atp_free    ( void *) RTRT_THROWING_EXCEPTION;
ATL_NAMESPACE_END
#undef ATP_LIBC_PROTO_PREFIX
#define ATP_LIBC_PROTO_PREFIX ATL_LIBC_PROTO_PREFIX
#endif /* RTRT_TDP_COMPILATION */
#endif /* ( RTRT_HEAP == RTRT_STD ) && ( RTRT_LOWLEVEL_MEMLIB == RTRT_NONE ) */

#if ( RTRT_HEAP == RTRT_STD ) && ( RTRT_LOWLEVEL_MEMLIB == RTRT_USR )
/* This is the link way to catch the calls */
#define RTRT_DO_MALLOC     lowlevel_malloc
#define RTRT_DO_CALLOC     lowlevel_calloc
#define RTRT_DO_REALLOC    lowlevel_realloc
#define RTRT_DO_FREE       lowlevel_free
#endif

#if RTRT_HEAP == RTRT_USR
/* This is the user way */
#define RTRT_DO_MALLOC     usr_malloc
#define RTRT_DO_FREE       usr_free
#endif

/*******************************************************************************
* Functions prototypes:
*******************************************************************************/

#if !RTRT_KR

_ATL_EXTERN_C                         int  _atp_dump                ( int );
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void _atp_link_parse_globals  ( _AtpGlobalsFuncTypeP );
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void _atp_check_var          ( char *,
                                                  RTRT_U_INT32);
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void *_atp_alloc( RTRT_SIZE_T size,  RTRT_U_INT16 Mtype);
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void _atp_freeT( void *ptr,   RTRT_U_INT16 Mtype);
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void _atp_start_test_driver   ( char * );
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void atp_early_init           ( void );
#ifdef _ATL_EXIT_PENDING
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void atp_exit_pending         ( void );
#else /* _ATL_EXIT_PENDING */
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void atp_late_exit            ( void );
#endif /* _ATL_EXIT_PENDING */
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void atl_dump_CALLSTACK       ( char *name, _AtlStackLinkP stack_link );
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     char *_PurifyLTHeapTrack 		( char * ptr, RTRT_SIZE_T s);

_ATL_EXTERN_LIBC ATP_LIBC_PROTO_PREFIX void * ATL_CALLCONV RTRT_DO_MALLOC  ( RTRT_SIZE_T );
_ATL_EXTERN_LIBC ATP_LIBC_PROTO_PREFIX void * ATL_CALLCONV RTRT_DO_CALLOC  ( RTRT_SIZE_T,
                                                            RTRT_SIZE_T );
_ATL_EXTERN_LIBC ATP_LIBC_PROTO_PREFIX void * ATL_CALLCONV RTRT_DO_REALLOC ( void *,
                                                            RTRT_SIZE_T );
_ATL_EXTERN_LIBC ATP_LIBC_PROTO_PREFIX void   ATL_CALLCONV RTRT_DO_FREE    ( void *);


#else /* RTRT_KR */

_ATL_EXTERN_C                         int  _atp_dump                ( );
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void _atp_link_parse_globals  ( );
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void _atp_check_var           ( );
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void *_atp_alloc 				( );
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void _atp_freeT				( );
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void _atp_start_test_driver   ( );
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void atp_early_init           ( );
#ifdef _ATL_EXIT_PENDING
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void atp_exit_pending         ( );
#else /* _ATL_EXIT_PENDING */
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void atp_late_exit            ( );
#endif /* _ATL_EXIT_PENDING */
_ATL_EXTERN_C ATL_TP_PROTO_PREFIX     void atl_dump_CALLSTACK  ( );
_ATL_EXTERN_LIBC ATP_LIBC_PROTO_PREFIX void *RTRT_DO_MALLOC  ( );
_ATL_EXTERN_LIBC ATP_LIBC_PROTO_PREFIX void *RTRT_DO_CALLOC  ( );
_ATL_EXTERN_LIBC ATP_LIBC_PROTO_PREFIX void *RTRT_DO_REALLOC ( );
_ATL_EXTERN_LIBC ATP_LIBC_PROTO_PREFIX void  RTRT_DO_FREE    ( );

#endif /* RTRT_KR */


/*******************************************************************************
* C++ Global new and delete definitions to ensure that they use malloc family
*******************************************************************************/

#ifdef RTRT_TDP_COMPILATION

/* User Heap API Types and Routines: */
#include "atpapi.h"

#if RTRT_CPLUSPLUS_LANGUAGE
#ifdef RTRT_DEFINE_NEW_DELETE
RTRT_DEFINE_NEW_DELETE
#endif /* RTRT_DEFINE_NEW_DELETE */
#endif /* RTRT_CPLUSPLUS_LANGUAGE */

#else /* RTRT_TDP_COMPILATION */

#if RTRT_CPLUSPLUS_LANGUAGE
#ifdef RTRT_INLINE_NEW_DELETE
RTRT_INLINE_NEW_DELETE
#endif /* RTRT_DEFINE_NEW_DELETE */
#endif /* RTRT_CPLUSPLUS_LANGUAGE */
#endif /* RTRT_TDP_COMPILATION */

#ifdef _ATP_PARSE_GLOBALS_NEEDED
#if _ATP_PARSE_GLOBALS_NEEDED
#if !RTRT_KR
static void _ATP_PARSE_GLOBALS ( void );
#else
static void _ATP_PARSE_GLOBALS ( );
#endif
#endif
#endif

#ifdef RTRT_TDP_COMPILATION
/*******************************************************************************
* Function prototype declaration for TP compilation
*******************************************************************************/
#if RTRT_64_BITS_PTR == 0
#define atp_dump_ptr(v) _atp_dump_uint((RTRT_U_INT32)(v))
#else
#define atp_dump_ptr(v) _atp_dump_ptr((RTRT_INT_SIZE_PTR)(v))
RTRT_STATICFCT void _atp_dump_ptr(
#if !RTRT_KR
	RTRT_INT_SIZE_PTR v
#endif
	);
#endif /* RTRT_64_BITS_PTR */

RTRT_STATICFCT void _atp_dump_uint(
#if !RTRT_KR
 RTRT_U_INT32 v
#endif
	);
RTRT_STATICFCT void _atp_dump_start_test();
RTRT_STATICFCT void _atp_stack_dump(
#if !RTRT_KR
_AtlStackLinkP stack_link, int call_stack_size
#endif
);
RTRT_STATICFCT void _atp_dump_err
#if !RTRT_KR
(const char * err,RTRT_INT_SIZE_PTR v);
#else
( err,v );
#endif

#endif /* RTRT_TDP_COMPILATION */
#endif /* ATP_H_INCLUDED */

