

/*
 *  Licensed Materials - Property of IBM
 *  Restricted Material of IBM
 *  IBM Rational Test RealTime
 *  (c) Copyright IBM Corp. 2003-2011 All Rights Reserved.
 *
 *  US Governement Users Restricted Rights - Use, duplication or
 *  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */


/*******************************************************************************
* $Descr: Call Stack Management (included by TP.c[pp]) $
*******************************************************************************/
#ifndef ATL_DUMP_ON_RETURNING
#define ATL_DUMP_ON_RETURNING
#endif

/*******************************************************************************
* Global variables
*******************************************************************************/

#if ATL_STACK_CALLING_LEVEL
RTRT_GLOBALVAR RTRT_U_INT16 _atl_stack_level_max;
#endif /* ATL_STACK_CALLING_LEVEL */

RTRT_GLOBALVAR RTRT_U_INT8 _atl_stack_error_emitted;
#ifndef RTRT_NB_SET_JMP
#define RTRT_NB_SET_JMP 10
#endif
typedef struct {
	void * env;
	_AtlStackLinkP stack_link;
}atl_set_jmp_table_t;
atl_set_jmp_table_t _atl_set_jmp_table[RTRT_NB_SET_JMP];

#if ATL_MULTI_THREADS == ATL_NO

RTRT_GLOBALVAR _AtlStackLinkP _atl_stack_start_link;
RTRT_GLOBALVAR _AtlStackLinkP _atl_stack_current_link;
#define _atl_get_stack_current() _atl_stack_current_link
#if ATL_STACK_CALLING_LEVEL
RTRT_GLOBALVAR RTRT_U_INT16 _atl_stack_level;
#endif /* ATL_STACK_CALLING_LEVEL */
#if USE_ATP && ATP_STACK_SIZE_LEVEL
RTRT_GLOBALVAR RTRT_U_INT16 _atl_stack_size_max_level;
RTRT_GLOBALVAR RTRT_U_INT32 _atl_stack_size_max;
RTRT_STATICVAR char	_atl_in_long_jmp;
#endif /* USE_ATP && ATP_STACK_SIZE_LEVEL */

#else /* ATL_MULTI_THREADS */

RTRT_GLOBALVAR _AtlStackLinkP _atl_stack_table_start[ATL_THREADS_MAX+1];
RTRT_GLOBALVAR _AtlStackLinkP _atl_stack_table_current[ATL_THREADS_MAX+1];

/*******************************************************************************
*  Service: _atl_get_stack_current
*******************************************************************************/
RTRT_STATICFCT _AtlStackLinkP _atl_get_stack_current
#if !RTRT_KR
( void )
#else
( )
#endif
{
  RTRT_THIDX_T thedata;
  RTRT_MT_THREAD_SELF_T self=RTRT_MT_THREAD_SELF;

  _AtlStackLinkP current=(_AtlStackLinkP)0; /* Not observed */
  RTRT_MT_THREAD_SPECIFIC_DATA_GET ( thedata );
  if ( thedata >= 0 &&
       thedata <= ATL_THREADS_MAX &&
       _atl_threads_table_used[thedata]==self )
  { /* In order to be sure this is something we are already observing */
  	current = (_AtlStackLinkP)_atl_stack_table_current[thedata];
#if ATL_CHECK_STACK
    if ( current && _ATL_STACK_CHECK_MAGIC(current) )
    {
      ATL_ERROR ('H', "_atl_get_stack_current(): runtime analysis stack data is corrupted" );
      return (_AtlStackLinkP)0;
    }
#endif /* ATL_CHECK_STACK */
  }

  return current;
}

#if ATL_STACK_CALLING_LEVEL
RTRT_GLOBALVAR RTRT_U_INT16 _atl_stack_table_level[ATL_THREADS_MAX+1];
#endif /* ATL_STACK_CALLING_LEVEL */
#if USE_ATP && ATP_STACK_SIZE_LEVEL
RTRT_GLOBALVAR RTRT_U_INT16 _atl_stack_table_size_max_level[ATL_THREADS_MAX+1];
RTRT_GLOBALVAR RTRT_U_INT32 _atl_stack_table_size_max[ATL_THREADS_MAX+1];
RTRT_STATICVAR char	_atl_in_long_jmp;
#endif /* USE_ATP && ATP_STACK_SIZE_LEVEL */

#define _atl_stack_start_link \
  _atl_stack_table_start[thread_index]
#define _atl_stack_current_link \
  _atl_stack_table_current[thread_index]

#if ATL_STACK_CALLING_LEVEL
#define _atl_stack_level \
  _atl_stack_table_level[thread_index]
#endif /* ATL_STACK_CALLING_LEVEL */
#if USE_ATP && ATP_STACK_SIZE_LEVEL
#define  _atl_stack_size_max_level \
 _atl_stack_table_size_max_level[thread_index]
#define  _atl_stack_size_max \
 _atl_stack_table_size_max[thread_index]
#endif /* USE_ATP && ATP_STACK_SIZE_LEVEL */

RTRT_MT_CRITICAL_VAR_DECL(atl_stack_critical)


/*******************************************************************************
*  Service: _atl_stack_check_thread
*******************************************************************************/
RTRT_STATICFCT RTRT_THIDX_T _atl_stack_check_thread
#if !RTRT_KR
( _AtlStackLinkP stack_link )
#else
( stack_link )
  _AtlStackLinkP stack_link;
#endif
{
  int thread_status;
  RTRT_THIDX_T thread_index=stack_link->ThreadIndex;
  thread_status = _atl_threads_get_status (&thread_index);
  if ( thread_status == _ATL_THREADS_CREATED )
  {
    _atl_stack_start_link = (_AtlStackLinkP)0;
    _atl_stack_current_link = (_AtlStackLinkP)0;
#if ATL_STACK_CALLING_LEVEL
    _atl_stack_level = 0;
#endif /* ATL_STACK_CALLING_LEVEL */
#if USE_ATP && ATP_STACK_SIZE_LEVEL
  _atl_stack_size_max_level =0;
  _atl_stack_size_max =0;
#endif /* USE_ATP && ATP_STACK_SIZE_LEVEL */
#if USE_ATT
  RTRT_MT_CRITICAL_VAR_LOCK(atl_tdf_mutex)
    _att_thread_fall_asleep ( );
    _att_thread_create ( stack_link, thread_index);
  RTRT_MT_CRITICAL_VAR_UNLOCK(atl_tdf_mutex)
#endif /* USE_ATT */
  }
  return thread_index;
}

#endif /* ATL_MULTI_THREADS */

#if RTRT_LONGJMP!=RTRT_NONE
/*******************************************************************************
*  Service: _atl_stack_jump
* idx = 0 when stackPointer is going UP, 1 Down,
*       2 when addr is not the stackPointer but a value used in the setjmp custom
*******************************************************************************/
_ATL_C_GLOBALFCT void _atl_stack_jump
#if !RTRT_KR
( void *addr, short idx )
#else
( addr, idx )
  void *addr;
  short idx;
#endif
{
  _AtlStackLinkP current,start;

#if ATL_MULTI_THREADS
  RTRT_THIDX_T thread_index;
  thread_index=_atl_threads_get_index ( );
#endif /* ATL_MULTI_THREADS */
  current = _atl_stack_current_link;
  if(idx >1)
  { /* new longjmp custom from 7.5 */

	for(idx=0; (idx<RTRT_NB_SET_JMP) && (_atl_set_jmp_table[idx].env!=addr); idx++);
	if(idx<RTRT_NB_SET_JMP)
	{
		start=_atl_set_jmp_table[idx].stack_link;
	}
	else
	{ /* keep only the first level if addr is not found in the table */
	  /* then let the pop re recompute it when the next methode returns */
		start = _atl_stack_start_link->Down;
	}
	for ( ;
			current &&
#if ATL_CHECK_STACK
			! _ATL_STACK_CHECK_MAGIC(current) &&
#endif
        	current != start;
	        current = current->Up )
	{
#if USE_ATQ
			if(current->AtqData)
				_atq_end ( current );
#endif /* USE_ATQ */
#if USE_ATT
			_att_method_return ( current );
#endif /* USE_ATT */
			_atl_stack_link_pop ( current );
#if USE_ATP && ATP_STACK_SIZE_LEVEL
			/* do record stack size only on the deepest while poping */
			_atl_in_long_jmp=1;
#endif /* USE_ATP && ATP_STACK_SIZE_LEVEL */
	}
#if USE_ATP && ATP_STACK_SIZE_LEVEL
	_atl_in_long_jmp=0;
#endif /* USE_ATP && ATP_STACK_SIZE_LEVEL */
  }
#if RTRT_LONGJMP==RTRT_USR
  else
  { /* old customization addr=stack pointer */

	  start = _atl_stack_start_link;

	  if(( idx ? ( addr > (void *)current ) :
							 ( (void *)current > addr ) )
	   &&(( idx ? ( addr < (void *)start ) :
							 ( (void *)start < addr ))))
	  {

		  for ( ;
			current &&
#if ATL_CHECK_STACK
			! _ATL_STACK_CHECK_MAGIC(current) &&
#endif
			( idx ? ( addr > (void *)current ) :
							 ( (void *)current > addr ) );
			current = current->Up )
		  {
#if USE_ATT
			_att_method_return ( current );
#endif /* USE_ATT */
#if USE_ATQ
			if(current->AtqData)
				_atq_end ( current );
#endif /* USE_ATQ */
			_atl_stack_link_pop ( current );
#if USE_ATP && ATP_STACK_SIZE_LEVEL
			/* do record stack size only on the deepest while poping */
			_atl_in_long_jmp=1;
#endif /* USE_ATP && ATP_STACK_SIZE_LEVEL */
		 }
#if USE_ATP && ATP_STACK_SIZE_LEVEL
	     _atl_in_long_jmp=0;
#endif /* USE_ATP && ATP_STACK_SIZE_LEVEL */
	  }
	  else
	  {
		  /* the selected register does not provide the stack pointer */
		if(addr)
		{
			ATL_ERROR ('A', "longjmp customization error" );
		}
		for ( ;
				current
#if ATL_CHECK_STACK
				&& ! _ATL_STACK_CHECK_MAGIC(current)
#endif
				; current = current->Up )
		{
#if USE_ATQ
		    if(current->AtqData)
			_atq_end ( current );
#endif /* USE_ATQ */
		}
	  }
  }
#endif /* RTRT_LONGJMP==RTRT_USR */
}

/* function available from code */
_ATL_C_GLOBALFCT void _atl_ext_stack_jump
#if !RTRT_KR
( void *addr, short idx )
#else
( addr, idx )
  void *addr;
  short idx;
#endif
{
	_atl_stack_jump(addr,idx);
}

/*******************************************************************************
*  Service: _atl_stack_set used in the new longJmp custom
*  sl is the stack_link where setjmp is executed
*  env is the value used in the longjmp
* this method memorize the stack_link of the function that longjmp targets
*******************************************************************************/
_ATL_C_GLOBALFCT void _atl_stack_set
#if !RTRT_KR
( void* sl ,void* env )
#else
( sl, env )
  void* sl;
  void* env;
#endif
{

int i;
	for(i=0; (i<RTRT_NB_SET_JMP) && (_atl_set_jmp_table[i].env); i++);
	if(i>=RTRT_NB_SET_JMP)
	{
		ATL_ERROR ('L', "longjmp table overflow re-define RTRT_NB_SET_JMP" );
		i=0; /* rollover */
	}
	_atl_set_jmp_table[i].env=env;
	_atl_set_jmp_table[i].stack_link=(_AtlStackLinkP)sl;
	i++;
	if(i<RTRT_NB_SET_JMP)_atl_set_jmp_table[i].env=0;

}

/* function available from code */
_ATL_C_GLOBALFCT void _atl_ext_stack_set
#if !RTRT_KR
( void* sl ,void* env )
#else
( sl, env )
  void* sl;
  void* env;
#endif
{
	_atl_stack_set(sl,env);
}

#endif /* RTRT_LONGJMP!=RTRT_NONE */

#if ATL_MULTI_THREADS

/*******************************************************************************
*  Service: _atl_stack_finish
*******************************************************************************/
RTRT_STATICFCT void _atl_stack_finish
#if !RTRT_KR
( RTRT_THIDX_T index )
#else
( index )
 RTRT_THIDX_T index;
#endif
{
  _AtlStackLinkP current;

  current = _atl_stack_table_current[index];
  for ( ;
        current
#if ATL_CHECK_STACK
        && ! _ATL_STACK_CHECK_MAGIC(current)
#endif
                                         ;
        current = current->Up )
  {
#if USE_ATT
    _att_method_return ( current );
#endif /* USE_ATT */
#if USE_ATQ
    if(current->AtqData)
       _atq_end ( current );
#endif /* USE_ATQ */
    _atl_stack_link_pop ( current );
#if USE_ATP && ATP_STACK_SIZE_LEVEL
			/* do record stack size only on the deepest while poping */
			_atl_in_long_jmp=1;
#endif /* USE_ATP && ATP_STACK_SIZE_LEVEL */
  }
#if USE_ATP && ATP_STACK_SIZE_LEVEL
  _atl_in_long_jmp=0;
#endif /* USE_ATP && ATP_STACK_SIZE_LEVEL */
}

#endif /* ATL_MULTI_THREADS */

/*******************************************************************************
*  Service: _atl_stack_link_push
*******************************************************************************/
_ATL_C_GLOBALFCT void _atl_stack_link_push
#if !RTRT_KR
( _AtlStackLinkP stack_link )
#else
( stack_link )
  _AtlStackLinkP stack_link;
#endif
{

#if ATL_MULTI_THREADS
  RTRT_THIDX_T	thread_index;
#endif /* ATL_MULTI_THREADS */
  if ( ! _ATL_INITIALIZED )
  {
    /* This branch is reached once in the process life */
    atl_early_init ( );
  }
#if ATL_CHECK_STACK
  _ATL_STACK_SET_MAGIC(stack_link->Magic1)
  _ATL_STACK_SET_MAGIC(stack_link->Magic2)
#endif /* ATL_CHECK_STACK */
  _ATL_STACK_INIT_ATP(stack_link)
  _ATL_STACK_INIT_ATQ(stack_link)

  if ( stack_link->Info.Source->UseCheck != _ATL_STACK_CONST &&
       ! _atl_stack_error_emitted )
  {
    ATL_ERROR ('J', "Observation Tools: Mismatch between tools selection for TDP compilation and for instrumentation; check your settings and command-line (Trace/Memory Profiling/Performance Profiling)" );
    _atl_stack_error_emitted = 1;
  }
#if ATL_MULTI_THREADS
  stack_link->ThreadIndex=0;
  thread_index=_atl_stack_check_thread ( stack_link );
  stack_link->ThreadIndex=thread_index;
#endif /* ATL_MULTI_THREADS */
  RTRT_MT_CRITICAL_VAR_LOCK(atl_stack_critical)
  if ( _atl_stack_start_link && _atl_stack_current_link
       /* Both are supposed to be null or non-null at the same time, but... */
#if ATL_CHECK_STACK
  		&& !(_ATL_STACK_CHECK_MAGIC(_atl_stack_current_link))
#endif /* ATL_CHECK_STACK */
  		)
  { /* Continue the stack */
    _atl_stack_current_link->Down = stack_link;
    stack_link->Up = _atl_stack_current_link;
  }
  else
  { /* Starting the stack */
    _atl_stack_start_link = stack_link;
    stack_link->Up = (_AtlStackLinkP)0;
  }
  stack_link->Down = (_AtlStackLinkP)0;
  _atl_stack_current_link = stack_link;
#if ATL_STACK_CALLING_LEVEL
  _atl_stack_level++;
  if ( _atl_stack_level_max < _atl_stack_level )
  {
    _atl_stack_level_max = _atl_stack_level;
  }
#endif /* ATL_STACK_CALLING_LEVEL */
  RTRT_MT_CRITICAL_VAR_UNLOCK(atl_stack_critical)
}

/*******************************************************************************
*  Service: _atl_stack_link_pop
*******************************************************************************/
_ATL_C_GLOBALFCT void _atl_stack_link_pop
#if !RTRT_KR
( _AtlStackLinkP stack_link )
#else
( stack_link )
  _AtlStackLinkP stack_link;
#endif
{
#if ATL_MULTI_THREADS
  RTRT_THIDX_T	thread_index;
#endif /* ATL_MULTI_THREADS */
  _AtlStackLinkP current;
  ATL_DUMP_ON_RETURNING

#if ATL_CHECK_STACK
  if ( stack_link && _ATL_STACK_CHECK_MAGIC(stack_link) )
  {
    ATL_ERROR ('K', "_atl_stack_link_pop(): runtime analysis stack data is corrupted" );
    return;
  }
#endif /* ATL_CHECK_STACK */
#if ATL_MULTI_THREADS
/* no need to call _atl_stack_check_thread ( stack_link ); since att did it */
  if(stack_link) thread_index=stack_link->ThreadIndex;
  else thread_index=_atl_threads_get_index();
  RTRT_MT_CRITICAL_VAR_LOCK(atl_stack_critical)
#endif /* ATL_MULTI_THREADS */
#if USE_ATP && ATP_STACK_SIZE_LEVEL
  if(!_atl_in_long_jmp)
  {
	/* stack actual size = start stack - actual stack pointer (parameter) */
#if RTRT_64_BITS_PTR ==0
	RTRT_U_INT32 end  =RTRT_PTR_TO_INT(&stack_link);
#else
	RTRT_U_INT64 end  =RTRT_PTR_TO_INT(&stack_link);
#endif
	RTRT_INT32 size =(RTRT_INT32)(RTRT_PTR_TO_INT(_atl_stack_start_link) - end);
	/* correct stack management UP/Down */
	if(size <0) size = -size;
    /* memorize only if the application uses more */
	if(size - _atl_stack_level*sizeof(_AtlStackLink) > _atl_stack_size_max)
	{
		_atl_stack_size_max=size - _atl_stack_level*sizeof(_AtlStackLink);
		_atl_stack_size_max_level=_atl_stack_level;
#if USE_ATT
	    _atl_tdf_line(_ATO_STYLE_COMMENT);
	    _att_print_note_text((char*)RTRT_CONST_STRING("StackMax detected"), stack_link->InstanceId );
#endif
		_atp_dump_start_test ( );
		/* Dump the Error event line */
		ATL_LINE_ADD_OPCODE_SPACE ( ATL_TPF_PE );
		RTRT_PRINT_ERROR_BEGIN
		atl_line_add ( "STKS" );
		/* print the stack pointer */
	    atp_dump_ptr(end);
		/* print the size */
		_atp_dump_uint(size);
		/* print the RTRT usage */
		_atp_dump_uint(_atl_stack_level*sizeof(_AtlStackLink));
		/* print the depth */
		_atp_dump_uint(_atl_stack_level);
#if ATL_MULTI_THREADS
		_atp_dump_uint(stack_link->ThreadIndex);
#endif /* ATL_MULTI_THREADS */


		RTRT_PRINT_ERROR
		atl_line_dump ( );

		/* Dump the current call stack with selected depth */
		_atp_stack_dump(stack_link,ATP_STACK_SIZE_LEVEL);
	}
  }
#endif /* USE_ATP && ATP_STACK_SIZE_LEVEL */
  if ( stack_link == _atl_stack_current_link )
  {
    if ( _atl_stack_start_link != _atl_stack_current_link )
    {
      _atl_stack_current_link = _atl_stack_current_link->Up;
      _atl_stack_current_link->Down = (_AtlStackLinkP)0;
    }
    else
    { /* Reaching the top of the stack */
      _atl_stack_start_link = (_AtlStackLinkP)0;
      _atl_stack_current_link = (_AtlStackLinkP)0;
    }
#if ATL_STACK_CALLING_LEVEL
    _atl_stack_level--;
#endif /* ATL_STACK_CALLING_LEVEL */
  }
  else
  { /* longjmp ? */
#if ATL_STACK_CALLING_LEVEL
    _atl_stack_level = 0;
#endif /* ATL_STACK_CALLING_LEVEL */
    for ( current = _atl_stack_start_link;
          current && current != stack_link
#if ATL_CHECK_STACK
    && !_ATL_STACK_CHECK_MAGIC(current)
#endif
          ;
          current = current->Down )
    {
#if ATL_STACK_CALLING_LEVEL
      _atl_stack_level++;
#endif /* ATL_STACK_CALLING_LEVEL */
    }
    if ( current && current == stack_link)
    {
      _atl_stack_current_link = stack_link->Up;
      if ( _atl_stack_current_link )
      {
        _atl_stack_current_link->Down = (_AtlStackLinkP)0;
      }
    }
    else
    { /* We are not in the same stack !?!
       * or this is a stack flush due to an exit() */
      _atl_stack_start_link = (_AtlStackLinkP)0;
      _atl_stack_current_link = (_AtlStackLinkP)0;
    }
  }
  RTRT_MT_CRITICAL_VAR_UNLOCK(atl_stack_critical)
}

/*******************************************************************************
*  Service: _atl_stack_link_pop_return
*******************************************************************************/
_ATL_C_GLOBALFCT int _atl_stack_link_pop_return
#if !RTRT_KR
( _AtlStackLinkP stack_link,
  int result )
#else
( stack_link, result )
  _AtlStackLinkP stack_link;
  int result;
#endif
{
  _atl_stack_link_pop ( stack_link );

  /* Must return result */
  return result;
}

/*******************************************************************************
*  Service: atl_stack_early_init
*******************************************************************************/
RTRT_STATICFCT void atl_stack_early_init
#if !RTRT_KR
( void )
#else
( )
#endif
{
#if ! RTRT_RAM_CLEAN

#if ATL_MULTI_THREADS
  RTRT_THIDX_T thread_index;
  for ( thread_index = 0; thread_index <= ATL_THREADS_MAX; thread_index++ )
#endif /* ATL_MULTI_THREADS */
  {
	  _atl_stack_start_link = (_AtlStackLinkP)0;
	  _atl_stack_current_link = (_AtlStackLinkP)0;
#if ATL_STACK_CALLING_LEVEL
  	  _atl_stack_level = 0;
#endif /* ATL_STACK_CALLING_LEVEL */
#if USE_ATP && ATP_STACK_SIZE_LEVEL
  _atl_stack_size_max_level =0;
  _atl_stack_size_max =0;
  _atl_in_long_jmp=0;
#endif /* USE_ATP && ATP_STACK_SIZE_LEVEL */
  }
#if ATL_STACK_CALLING_LEVEL
  _atl_stack_level_max = 0;
#endif /* ATL_STACK_CALLING_LEVEL */
  _atl_stack_error_emitted = 0;
  _atl_set_jmp_table[0].env = 0;
#endif

  RTRT_MT_CRITICAL_VAR_INIT(atl_stack_critical) /* mutex protecting the call stack manipulations */
}

#ifdef _ATL_EXIT_PENDING
/*******************************************************************************
*  Service: atl_stack_exit_pending
*******************************************************************************/
_ATL_C_GLOBALFCT void atl_stack_exit_pending
#if !RTRT_KR
( void )
#else
( )
#endif
{
 /* _atl_stack_link_pop ( (_AtlStackLinkP)0 ); */
}
#else
/*******************************************************************************
*  Service: atl_stack_clean_up
*******************************************************************************/
_ATL_C_GLOBALFCT void atl_stack_clean_up
#if !RTRT_KR
( void )
#else
( )
#endif
{
  /* _atl_stack_link_pop ( (_AtlStackLinkP)0 ); */
}
/*******************************************************************************
*  Service: atl_stack_late_exit
*******************************************************************************/
_ATL_C_GLOBALFCT void atl_stack_late_exit
#if !RTRT_KR
( void )
#else
( )
#endif
{
  RTRT_MT_CRITICAL_VAR_DESTROY(atl_stack_critical)
}
#endif /* _ATL_EXIT_PENDING */

